package com.sratest.proxy;

import java.io.Serializable;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;

public class GenericDTOImpl implements InvocationHandler, Serializable {
	private static final long                   serialVersionUID = -1381186792996519168L;
	private              HashMap<String,Object> values           = new HashMap<String,Object>();
	private 			 Boolean				inHashCodeSync	 = false;
	private 			 Boolean				inEqualsSync	 = false;

	public GenericDTOImpl(HashMap<String,Object> values) {
		this.values = values;
	}

	public Object invoke(Object proxy, Method m, Object[] args) {
		String methodName     = m.getName();
    	Class  declaringClass = m.getDeclaringClass();

    	if (declaringClass == Object.class) {
    		if ("hashCode".equals(methodName)) {
    			synchronized( inHashCodeSync ) {
    				if( inHashCodeSync ) {
    					return 0;
    				}
    				try {
	    				inHashCodeSync = true;
	        			// Use the hash of the underlying value map
	        			return this.values.hashCode();
    				} finally {
	        			inHashCodeSync = false;
    				}
    			}
    		}
    		if ("equals".equals(methodName)) {
    			// return proxyEquals(proxy, args[0]);

    			// Equal if the other one is a proxy as well and has equal value map
    			if( args[0] != null && args[0] instanceof Proxy ) {
    				InvocationHandler handler = Proxy.getInvocationHandler( args[0] );
    				
    				if( handler instanceof GenericDTOImpl ) {
    	    			synchronized( inEqualsSync ) {
	        				if( inEqualsSync ) {
	        					return true;
	        				}
	        				try {
	        					inEqualsSync = true;
	        					return this.values.equals( ((GenericDTOImpl) handler).values );
	        				} finally {
	        					inEqualsSync = false;
	        				}
    	    			}
    				}
    			}

    			// Null, not a proxy or different handler impl class;
    			// theoretically, the other object could be semantically equal
    			// (all getters return value equal to the ones our proxy returns),
    			// if it is a proxy with different impl, or a "real" object. 
    			// But to check this, we'd need to compare all getter results, as
    			// we are totally ignorant about the implementation of the other
    			// object.
    			return false;
    		}
    		if ("toString".equals(methodName)) {
    			return proxyToString(proxy);
    		}
    		throw new InternalError(
    			    "unexpected Object method dispatched: " + m);
    	}
    	String fieldName = getFieldName(methodName);

    	if (methodName.startsWith("set")
    			&& this.values.containsKey(fieldName)) {
    		this.values.put(fieldName, args[0]);
    		return null;
    	}
    	if ((methodName.startsWith("get")
    			|| methodName.startsWith("is"))
    			&& this.values.containsKey(fieldName)) {
    		return this.values.get(fieldName);
    	}
    	throw new RuntimeException(new NoSuchMethodException(methodName));
    }
 
	protected Integer proxyHashCode(Object proxy) {
		return new Integer(System.identityHashCode(proxy));
	}

	protected Boolean proxyEquals(Object proxy, Object other) {
		return (proxy == other ? Boolean.TRUE : Boolean.FALSE);
	}

	protected String proxyToString(Object proxy) {
		return proxy.getClass().getName() + '@' +
		    Integer.toHexString(proxy.hashCode());
	}

	protected static String getFieldName(String methodName) {
		if (methodName.startsWith("is")) {
			return methodName.substring(2, 3).toLowerCase() + (methodName.length() > 3 ? methodName.substring(3) : "");
		}
		return methodName.substring(3, 4).toLowerCase() + (methodName.length() > 4 ? methodName.substring(4) : "");
	}

	public HashMap<String,Object> getValues() {
		return this.values;
	}
}
